/*
 * Copyright (c) Oona Räisänen
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
#ifndef TABLES_H_
#define TABLES_H_

#include <cstdint>
#include <string>
#include <string_view>

namespace redsea {

std::string_view getPTYNameString(std::uint16_t pty);
std::string_view getPTYNameStringRBDS(std::uint16_t pty);
std::string_view getCountryString(std::uint16_t cc, std::uint16_t ecc);
std::string_view getLanguageString(std::uint16_t code);
std::string_view getAppNameString(std::uint16_t aid);
std::string_view getRTPlusContentTypeString(std::uint32_t content_type);
std::string_view getDICodeString(std::uint16_t di);
std::string getCallsignFromPI(std::uint16_t pi);

}  // namespace redsea
#endif  // TABLES_H_
